import * as MC from '@minecraft/server';
import * as MCUI from '@minecraft/server-ui';

// The code is a mess right now, the full release version will be much cleaner.



function SplitString(str, parts) {
    let result = [];
    let partLength = Math.floor(str.length / parts);
    let remainder = str.length % parts;
    let startIndex = 0;
    for (let i = 0; i < parts; i++) {
        let length = partLength + (i < remainder ? 1 : 0);
        result.push(str.slice(startIndex, startIndex + length));
        startIndex += length;
    };
    return result;
};

function arrayEqual(arr1, arr2) {
    if (arr1.length !== arr2.length) { return false; };
    for (let i = 0; i < arr1.length; i++) {
        if (arr1[i] !== arr2[i]) { return false; };
    };
    return true;
};

function PosToLoc(pos) {
    return { 'x': pos[0], 'y': pos[1], 'z': pos[2] };
};
function LocToPos(loc) {
    return [loc.x, loc.y, loc.z];
};
function PosInArea(area, pos) {
    for (let i = 0; i < 3; i++) {
        if (area.start[i] <= area.end[i]) {
            if (pos[i] < area.start[i] || pos[i] > area.end[i]) {
                return false;
            };
        }
        else {
            if (pos[i] < area.end[i] || pos[i] > area.start[i]) {
                return false;
            };
        };
    };
    return true;
};
function OutlineArea(area, dimension, particleId) {
    let i = -1;
    area.start.forEach(coord => {
        i += 1;
        let coordInc = 1;
        if (area.end[i] < coord) { coordInc = -1; };
        for (let i_ = coord; i_ !== area.end[i] + coordInc; i_ += coordInc) {
            let position = [area.start[0], area.start[1], area.start[2]];
            position[i] = i_;
            let positionJoined = position.join(' ');
            try {
                dimension.runCommand(`particle ${particleId} ${positionJoined}`);
            } catch { };
        };
    });
    i = -1;
    area.end.forEach(coord => {
        i += 1;
        let coordInc = 1;
        if (area.start[i] < coord) { coordInc = -1; };
        for (let i_ = coord; i_ !== area.start[i] + coordInc; i_ += coordInc) {
            let position = [area.end[0], area.end[1], area.end[2]];
            position[i] = i_;
            let positionJoined = position.join(' ');
            try {
                dimension.runCommand(`particle ${particleId} ${positionJoined}`);
            } catch { };
        };
    });
};

function TranslateFromRelativePosition(TargetPos, FacingState) {
    if (FacingState === 'north') {
        return [TargetPos[0], TargetPos[1], TargetPos[2]];
    };
    if (FacingState === 'south') {
        return [TargetPos[0], TargetPos[1], -TargetPos[2]];
    };
    if (FacingState === 'west') {
        return [TargetPos[2], TargetPos[1], TargetPos[0]];
    };
    if (FacingState === 'east') {
        return [-TargetPos[2], TargetPos[1], TargetPos[0]];
    };
    if (FacingState === 'up') {
        return [TargetPos[0], -TargetPos[2], TargetPos[1]];
    };
    if (FacingState === 'down') {
        return [TargetPos[0], TargetPos[2], TargetPos[1]];
    };
    console.warn(`Unknown FacingState: ${FacingState}`);
};




// Data APIs //
///////////////
let internalData = {};
const inDapi = {
    'get': function inDapiGet(key) { return internalData[key]; },
    'set': function inDapiSet(key, value) { internalData[key] = value; }
};
const mcDapi = {
    'get': function mcDapiGet(key) {
        let result = [];
        let i = -1;
        while (true) {
            i += 1;
            let x = MC.world.getDynamicProperty(`${key}[${i}]`);
            if (x === undefined) { break; }
            else { result.push(x); };
        };
        if (result.length === 0) { return undefined; };
        console.warn(result.join(''));
        return JSON.parse(result.join(''));
    },
    'set': function mcDapiSet(key, value) {
        if (typeof value !== 'object') {
            console.warn(`ERROR mcDapi.set: "value" should be an object.`);
            return;
        };
        let stringValue = JSON.stringify(value);
        //dynamic properties must be split when they are large enough due to the limits imposed by the Minecraft Server API.
        let parts = Math.round(stringValue.length / 9500);
        if (parts < 1) { parts = 1; };
        let splitValue = SplitString(stringValue, parts);
        for (let i = 0; i < parts; i++) {
            MC.world.setDynamicProperty(`${key}[${i}]`, splitValue[i]);
        };
    }
};


// Setup Grid Data
if (mcDapi.get('GRIDS') === undefined) {
    let grids = {
        'all': {},
        'curId': 0,
    };
    mcDapi.set('GRIDS', grids);
};
var GRIDS = mcDapi.get('GRIDS');
var GridTerminalTypes = ['gw:grid_terminal'];
var AreaEditors = {};
function SYNC() {
    mcDapi.set('GRIDS', GRIDS);
};
function NewGridId() {
    GRIDS.curId += 1;
    return GRIDS.curId;
};
function GridTick(id) {
    let grid = GRIDS.all[id];
    let dimension = MC.world.getDimension(grid.dimension);
    let newGridVelo = [grid.velocity[0], grid.velocity[1], grid.velocity[2]];
    let newGridPos = [grid.position[0], grid.position[1], grid.position[2]];
    if (grid.config.gravity === true) { newGridVelo[1] -= 0.02; };
    if (grid.config.drag === true) {
        let i = -1;
        newGridVelo.forEach(value => {
            i += 1;
            if (value > 0) { newGridVelo[i] -= 0.005; };
            if (value < 0) { newGridVelo[i] += 0.005; };
        });
    };
    let oldGridPosRound = [Math.round(grid.position[0]), Math.round(grid.position[1]), Math.round(grid.position[2])];
    let i = -1;
    newGridVelo.forEach(value => {
        i += 1;
        newGridPos[i] += value;
    });
    let gridPosRound = [Math.round(newGridPos[0]), Math.round(newGridPos[1]), Math.round(newGridPos[2])];
    let gridPosDiff = [gridPosRound[0] - oldGridPosRound[0], gridPosRound[1] - oldGridPosRound[1], gridPosRound[2] - oldGridPosRound[2]];
    grid.hasUnloadedBlocks = false;
    let gridAlive = false;
    let moveGrid = true;
    let gridBlocksToUpdate = [];
    i = -1;
    grid.blocks.forEach(b => {
        i += 1;
        if (b === null || b === undefined) {
            grid.blocks[i] = null; return;
        };
        let block = dimension.getBlock(PosToLoc(b.position));
        if (block === undefined) {
            grid.hasUnloadedBlocks = true;
            return;
        };
        if (block.typeId !== b.type) {
            grid.blocks[i] = null;
            return;
        };
        if (GridTerminalTypes.includes(b.type) === true) {
            gridAlive = true;
            let facingState = block.permutation.getState('minecraft:facing_direction');
            function relativeMove(vec3) {
                vec3 = TranslateFromRelativePosition(vec3, facingState);
                newGridVelo[0] += vec3[0];
                newGridVelo[1] += vec3[1];
                newGridVelo[2] += vec3[2];
            };
            b.events.forEach(event => {
                let offsetBlock = block.offset(PosToLoc(TranslateFromRelativePosition(event.offset, facingState)));
                let powerLevel = null;
                if (offsetBlock === undefined) { powerLevel = 0 }
                else {
                    powerLevel = offsetBlock.getRedstonePower();
                    if (powerLevel === undefined) { powerLevel = 0 };
                };
                if (event.condition_redstone !== powerLevel && event.condition_redstone !== -1) { return };
                if (event.action === 'accel.forward') { relativeMove([0, 0, -event.accel_amount]); }
                else if (event.action === 'accel.backward') { relativeMove([0, 0, event.accel_amount]) }
                else if (event.action === 'accel.left') { relativeMove([-event.accel_amount, 0, 0]) }
                else if (event.action === 'accel.right') { relativeMove([event.accel_amount, 0, 0]) }
                else if (event.action === 'accel.up') { relativeMove([0, event.accel_amount, 0]) }
                else if (event.action === 'accel.down') { relativeMove([0, -event.accel_amount, 0]) };
            });
        };
        let newPos = [b.position[0], b.position[1], b.position[2]];
        newPos[0] += gridPosDiff[0]; newPos[1] += gridPosDiff[1]; newPos[2] += gridPosDiff[2];
        gridBlocksToUpdate.push([b, newPos]);
    });
    grid.blocks = grid.blocks.filter(item => { return item !== null });
    let gridAreasToUpdate = [];
    let commandsToRun = [];
    let commandsToRunAfter = [];
    let entitiesToMove = [];
    let rand = Math.round(Math.random() * 1000);
    if (grid.hasUnloadedBlocks === false && gridAlive === true) {
        i = -1;
        grid.areas.forEach(a => {
            i += 1;
            if (a === null || a === undefined) {
                grid.areas[i] = null;
                return;
            };
            if (a.start[1] > 320 || a.start[1] < -64) {
                grid.areas[i] = null;
                return;
            };
            if (a.end[1] > 320 || a.end[1] < -64) {
                grid.areas[i] = null;
                return;
            };
            let oldStart = PosToLoc(a.start); let oldEnd = PosToLoc(a.end);
            let newStart = Object.assign({}, oldStart); let newEnd = Object.assign({}, oldEnd);
            newStart.x += gridPosDiff[0]; newStart.y += gridPosDiff[1]; newStart.z += gridPosDiff[2];
            newEnd.x += gridPosDiff[0]; newEnd.y += gridPosDiff[1]; newEnd.z += gridPosDiff[2];
            let structId = `${grid.id}_${rand}_${i}`;
            let oldStartJoined = `${oldStart.x} ${oldStart.y} ${oldStart.z}`; let oldEndJoined = `${oldEnd.x} ${oldEnd.y} ${oldEnd.z}`;
            let newStartJoined = `${newStart.x} ${newStart.y} ${newStart.z}`; let newEndJoined = `${newEnd.x} ${newEnd.y} ${newEnd.z}`;
            if (oldStartJoined !== newStartJoined) {
                if (newStart.y > 320 || newStart.y < -64) {
                    moveGrid = false;
                    return;
                };
                if (newEnd.y > 320 || newEnd.y < -64) {
                    moveGrid = false;
                    return;
                };
                let xInc = 1; if (newStart.x > newEnd.x) { xInc = -1 };
                let yInc = 1; if (newStart.y > newEnd.y) { yInc = -1 };
                let zInc = 1; if (newStart.z > newEnd.z) { zInc = -1 };
                for (let x = newStart.x; x !== newEnd.x + xInc; x += xInc) {
                    for (let y = newStart.y; y !== newEnd.y + yInc; y += yInc) {
                        for (let z = newStart.z; z !== newEnd.z + zInc; z += zInc) {
                            let blockInGrid = false;
                            grid.areas.forEach(area => {
                                if (PosInArea(area, [x, y, z]) === true) {
                                    blockInGrid = true;
                                };
                            });
                            if (blockInGrid === true) { continue };
                            let loc = PosToLoc([x, y, z]);
                            let block = dimension.getBlock(loc);
                            if (block === undefined) {
                                moveGrid = false;
                                return;
                            };
                            if (block.isAir === false && block.isLiquid === false) {
                                moveGrid = false;
                                return;
                            };
                        };
                    };
                };
                gridAreasToUpdate.push([a, { 'start': LocToPos(newStart), 'end': LocToPos(newEnd) }]);
                commandsToRun.push(`structure save GRIDS.${structId} ${oldStartJoined} ${oldEndJoined} false memory true`);
                commandsToRun.push(`fill ${oldStartJoined} ${oldEndJoined} air`);
                let loadPos = LocToPos(newStart);
                if (newStart.x > newEnd.x) { loadPos[0] = newEnd.x };
                if (newStart.y > newEnd.y) { loadPos[1] = newEnd.y };
                if (newStart.z > newEnd.z) { loadPos[2] = newEnd.z };
                commandsToRunAfter.push(`structure load GRIDS.${structId} ${loadPos.join(' ')}`);
                let entities = dimension.getEntities();
                entities.forEach(entity => {
                    if (PosInArea(a, LocToPos(entity.location)) === true) {
                        entitiesToMove.push(entity);
                    };
                });
            };
        });
    };
    grid.areas = grid.areas.filter(item => { return item !== null });
    if (grid.hasUnloadedBlocks === true) {
        return;
    };
    if (gridAlive === false) {
        delete GRIDS.all[id];
        return;
    };
    if (moveGrid === true) {
        grid.position = newGridPos;
        grid.velocity = newGridVelo;
        let i = -1;
        grid.velocity.forEach(num => {
            i += 1;
            if (num > grid.config.maxVelo) { grid.velocity[i] = grid.config.maxVelo }
            else if (num < -grid.config.maxVelo) { grid.velocity[i] = -grid.config.maxVelo };
        });
        gridAreasToUpdate.forEach(pack => {
            // pack[0] = area Object - pack[1] = new area
            pack[0].start = pack[1].start; pack[0].end = pack[1].end;
        });
        gridBlocksToUpdate.forEach(pack => {
            pack[0].position = pack[1];
        });
        commandsToRun.forEach(cmd => {
            dimension.runCommand(cmd);
        });
        commandsToRunAfter.forEach(cmd => {
            dimension.runCommand(cmd);
        });
        entitiesToMove.forEach(entity => {
            let loc = entity.location;
            let newLoc = Object.assign({}, loc);
            newLoc.x += gridPosDiff[0]; newLoc.y += gridPosDiff[1]; newLoc.z += gridPosDiff[2];
            entity.tryTeleport(newLoc, { 'checkForBlocks': true });
        });
    }
    else {
        grid.velocity = [0, 0, 0];
    };
};



// Define UI-Forms //
/////////////////////
const FORMS = {
    'Message': (Player, Title, Message) => {
        let F = new MCUI.ActionFormData()
            .title(`${Title}`)
            .body(`${Message}`)
            .button('Okay');
        MC.system.run(() => {
            F.show(Player).then(Form => {
                return;
            });
        });
    },
    'GridTerminal-main': (Player, Grid, Terminal) => {
        if (Grid === undefined) {
            FORMS['Message'](Player, 'Grid Missing', `Control Panel got disconnected from its's Grid.`);
            return;
        };
        if (Grid.config.owner !== null && Grid.config.owner !== Player.name) {
            FORMS['Message'](Player, 'Grid Access Denied', `This Grid was set to private by §d${Grid.config.owner}§r.`);
            return;
        };
        let areaCount = Grid.areas.length;
        let blockCount = Grid.blocks.length;
        let gridUnloadedBlocksAlert = '';
        if (Grid.hasUnloadedBlocks === true) { gridUnloadedBlocksAlert = `\n§l§4ALERT: §rGrid contains block(s) that are out of range! The Grid cannot move while there are unloaded blocks.\n`; };
        let F = new MCUI.ActionFormData()
            .title('Control Panel')
            .body(`${gridUnloadedBlocksAlert}§dVelocity:§r ${Grid.velocity[0]}, ${Grid.velocity[1]}, ${Grid.velocity[2]}\n§dAreas:§r ${areaCount}\n§dTerminals:§r ${blockCount}`)
            .button('§3Basic Settings')
            .button('§3Area Editor')
            .button('§3Event Editor\n§8Use to set up controls');
        MC.system.run(() => {
            F.show(Player).then(Form => {
                if (Form.canceled) { return; };
                let choice = Form.selection;
                if (choice === 0) { FORMS['GridTerminal-settings'](Player, Grid, Terminal); }
                else if (choice === 1) { FORMS['GridTerminal-areaEditor'](Player, Grid, Terminal); }
                else if (choice === 2) { FORMS['GridTerminal-events'](Player, Grid, Terminal); };
            });
        });
    },
    'GridTerminal-settings': (Player, Grid, Terminal) => {
        let gc = Grid.config;
        let isPrivate = false;
        if (gc.owner !== null) { isPrivate = true; };
        let F = new MCUI.ModalFormData()
            .title('Control Panel - Settings')
            .toggle('Do Gravity?', gc.gravity)
            .toggle('Do Drag? §8(Naturally slows down over time)', gc.drag)
            .toggle('Private Access? §8(Only you can access this Grid\'s terminals)', isPrivate)
            .slider(`Max Velocity\n§8Currently: ${gc.maxVelo}§r`, 0, 1, 0.01, gc.maxVelo)
        MC.system.run(() => {
            F.show(Player).then(Form => {
                if (Form.canceled) {
                    FORMS['GridTerminal-main'](Player, Grid, Terminal);
                    return;
                };
                let Values = Form.formValues;
                gc.gravity = Values[0];
                gc.drag = Values[1];
                if (Values[2] === true) { gc.owner = Player.name; }//Player.name; }
                else { gc.owner = null; };
                gc.maxVelo = Values[3];
                FORMS['GridTerminal-main'](Player, Grid, Terminal);
            });
        });
    },
    'GridTerminal-areaEditor': (Player, Grid, Terminal) => {
        let F = new MCUI.ActionFormData()
            .title('Control Panel - Area Editor')
            .body('§6Adding an area:§r\nClick a block to set the starting corner, click again to set the ending corner. Once both corners are set the area will be saved.\n§6Deleting an area:§r\nClick on a block anywhere in the area.')
            .button('§l§2Add area')
            .button('§l§4Delete area')
        MC.system.run(() => {
            F.show(Player).then(Form => {
                if (Form.canceled === true) {
                    FORMS['GridTerminal-main'](Player, Grid, Terminal);
                    return;
                };
                let choice = Form.selection;
                if (choice === 0) {
                    AreaEditors[Player.name] = { 'grid': Grid, 'type': 'add', 'start': null, 'end': null };
                }
                else if (choice === 1) {
                    AreaEditors[Player.name] = { 'grid': Grid, 'type': 'delete' };
                };
            });
        });
    },
    'GridTerminal-events': (Player, Grid, Terminal) => {
        let F = new MCUI.ActionFormData()
            .title('Control Panel - Events')
            .body(`Configure panel events. Each panel has it's own events.`)
            .button('§l§2Add Event')
            .button('§l§4Delete Event');
        Terminal.events.forEach(event => {
            F.button(`§uEdit Event: §r${event.name}`);
        });
        MC.system.run(() => {
            F.show(Player).then(Form => {
                if (Form.canceled) {
                    FORMS['GridTerminal-main'](Player, Grid, Terminal);
                    return;
                };
                let choice = Form.selection;
                if (choice === 0) { // Add Event
                    Terminal.events.push({
                        'name': `New Event ${Terminal.events.length + 1}`,
                        'offset': [0, 0, -1],
                        'condition_redstone': -1,
                        'action': 'accel.forward',
                        'accel_amount': 0
                    });
                    FORMS['GridTerminal-events-editor'](Player, Grid, Terminal, Terminal.events.length - 1);
                    return;
                }
                else if (choice === 1) { // Delete Event
                    FORMS['GridTerminal-events-delete'](Player, Grid, Terminal);
                    return;
                }
                else { // Edit Event
                    FORMS['GridTerminal-events-editor'](Player, Grid, Terminal, choice - 2);
                    return;
                };
            });
        });
    },
    'GridTerminal-events-delete': (Player, Grid, Terminal) => {
        let eventsOptions = [];
        Terminal.events.forEach(event => {
            eventsOptions.push(event.name);
        });
        let F = new MCUI.ModalFormData()
            .title('Control Panel - Events - Delete')
            .dropdown('Select Event', eventsOptions, 0);
        MC.system.run(() => {
            F.show(Player).then(Form => {
                if (Form.canceled === true) {
                    FORMS['GridTerminal-events'](Player, Grid, Terminal);
                    return;
                };
                let values = Form.formValues;
                let i = -1;
                Terminal.events.forEach(event => {
                    i += 1;
                    if (i === values[0]) { Terminal.events[i] = null };
                });
                Terminal.events = Terminal.events.filter(item => { return item !== null });
                FORMS['GridTerminal-events'](Player, Grid, Terminal);
            });
        });
    },
    'GridTerminal-events-editor': (Player, Grid, Terminal, EventIndex = null) => {
        let event = Terminal.events[EventIndex];
        let eventAction = null;
        if (event.action === 'accel.forward') { eventAction = 0 }
        else if (event.action === 'accel.backward') { eventAction = 1 }
        else if (event.action === 'accel.left') { eventAction = 2 }
        else if (event.action === 'accel.right') { eventAction = 3 }
        else if (event.action === 'accel.up') { eventAction = 4 }
        else if (event.action === 'accel.down') { eventAction = 5 };
        let F = new MCUI.ModalFormData()
            .title('Control Panel - Events - Editor')
            .textField(`§uEvent Name`, '', `${event.name}`)
            .textField('§cCondition Location\n§8The location relative to the Control Panel to test the condition.', 'Example: "0 0 -1" (behind the panel)', event.offset.join(' '))
            .slider(`§cCondition:§r redstone signal\n§8(-1 = no condition)\nCurrently: ${event.condition_redstone}§r`, -1, 15, 1, event.condition_redstone)
            .dropdown(`§uAction`, ['Accel Forward', 'Accel Backward', 'Accel Left', 'Accel Right', 'Accel Up', 'Accel Down'], eventAction)
            .slider(`§uAccel Amount (Blocks per tick)\n§8Only for acceleration actions\nCurrently: ${event.accel_amount}§r`, 0, 1, 0.01, event.accel_amount)
        MC.system.run(() => {
            F.show(Player).then(Form => {
                if (Form.canceled === true) {
                    FORMS['GridTerminal-events'](Player, Grid, Terminal);
                    return;
                };
                let Values = Form.formValues;
                let eventOffset = Values[1].split(' ');
                function eventOffsetFail() {
                    FORMS['Message'](Player, 'Bad Input', 'Condition Offset should be a set of positional values.');
                }
                if (eventOffset.length !== 3) {
                    eventOffsetFail();
                    return;
                };
                let eventOffsetSuccess = false;
                try {
                    let i = -1;
                    eventOffset.forEach(num => {
                        i += 1;
                        eventOffset[i] = parseFloat(num);
                    });
                    eventOffsetSuccess = true;
                } catch { };
                if (eventOffsetSuccess === false) {
                    eventOffsetFail();
                    return;
                };
                event.name = Values[0];
                event.offset = eventOffset;
                event.condition_redstone = Values[2];
                if (Values[3] === 0) { event.action = 'accel.forward' }
                else if (Values[3] === 1) { event.action = 'accel.backward' }
                else if (Values[3] === 2) { event.action = 'accel.left' }
                else if (Values[3] === 3) { event.action = 'accel.right' }
                else if (Values[3] === 4) { event.action = 'accel.up' }
                else if (Values[3] === 5) { event.action = 'accel.down' };
                event.accel_amount = Values[4];
                FORMS['GridTerminal-events'](Player, Grid, Terminal);
            });
        });
    }
};




// Main Tick //
///////////////
var Timers = {
    'ticks': 0,
    'saveGrids': 0,
    'outlineAreas': 0,
    'playerInteractWithBlock': 0
};
function MainTick() {
    Object.keys(Timers).forEach(timer => { Timers[timer] += 1; });

    // Grid Tick
    Object.values(GRIDS.all).forEach(g => { GridTick(g.id); });

    // Area Editor
    Object.keys(AreaEditors).forEach(name => {
        let ae = AreaEditors[name];
        let player = MC.world.getPlayers({ 'name': name })[0];
        if (player === undefined) {
            delete AreaEditors[name];
            return;
        };
        if (player.isSneaking === true) {
            delete AreaEditors[name];
            return;
        };
        if (Timers.outlineAreas >= 3) {
            Timers.outlineAreas = 0;
            player.runCommand(`title @s actionbar §6Sneak to exit Area Editor`);
            let particle = 'gw:block';
            if (ae.type === 'add') {
                if (ae.start !== null) {
                    let lookingAtBlock = player.getBlockFromViewDirection({ 'maxDistance': 30 });
                    let lookingAtPos = null;
                    if (lookingAtBlock === undefined) { lookingAtPos = LocToPos(player.location) }
                    else { lookingAtPos = LocToPos(lookingAtBlock.block.location) };
                    let i = -1;
                    lookingAtPos.forEach(coord => {
                        i += 1;
                        lookingAtPos[i] = Math.round(coord);
                    });
                    let tooBig = false;
                    for (let i = 0; i < 3; i++) {
                        if (Math.abs(ae.start[i] - lookingAtPos[i]) >= 64) { tooBig = true };
                    };
                    if (tooBig === true) { particle = 'gw:block_red' };
                    OutlineArea({ 'start': ae.start, 'end': lookingAtPos }, MC.world.getDimension(ae.grid.dimension), particle);
                };
            }
            else if (ae.type === 'delete') {
                ae.grid.areas.forEach(area => {
                    OutlineArea(area, MC.world.getDimension(ae.grid.dimension), particle);
                });
            };
        };
    });

    //Save grid data to the world every 5 seconds (100 ticks)
    if (Timers.saveGrids >= 100) {
        Timers.saveGrids = 0;
        SYNC();
    };

    MC.system.run(MainTick);
};




// On Player Click Block // (before)
///////////////////////////
MC.world.beforeEvents.playerInteractWithBlock.subscribe(Event => {
    if (Timers.playerInteractWithBlock < 2) { return; }; // Prevent multiple executions in a single tick
    Timers.playerInteractWithBlock = 0;
    let blockPos = LocToPos(Event.block.location);
    if (GridTerminalTypes.includes(Event.block.typeId) === true) {
        let grid = undefined;
        let terminal = undefined;
        Object.keys(GRIDS.all).forEach(g => {
            GRIDS.all[g].blocks.forEach(b => {
                if (b === null || b === undefined) {
                    return;
                };
                if (arrayEqual(b.position, blockPos) === true) {
                    grid = GRIDS.all[g];
                    terminal = b;
                };
            });
        });
        FORMS['GridTerminal-main'](Event.player, grid, terminal);
    };

    Object.keys(AreaEditors).forEach(name => {
        let ae = AreaEditors[name];
        if (ae.type === 'add') {
            if (ae.start === null) {
                ae.start = blockPos;
            }
            else if (ae.end === null) {
                let tooBig = false;
                for (let i = 0; i < 3; i++) {
                    if (Math.abs(ae.start[i] - blockPos[i]) >= 64) { tooBig = true };
                };
                if (tooBig === true) {
                    Event.player.sendMessage(`§cAreas can't be longer than 64 blocks in any direction.`);
                }
                else {
                    ae.end = blockPos;
                };
            }
            else {
                ae.grid.areas.push({ 'start': ae.start, 'end': ae.end });
                delete AreaEditors[name];
                return;
            };
        }
        else if (ae.type === 'delete') {
            let i = -1;
            let stop = false;
            ae.grid.areas.forEach(area => {
                i += 1;
                if (stop === true) { return };
                if (PosInArea(area, blockPos) === true) {
                    ae.grid.areas[i] = null;
                    stop = true;
                    delete AreaEditors[name];
                };
            });
            ae.grid.areas = ae.grid.areas.filter(item => { return item !== null });
        };
    });
});


// On Entity Hit Block // (after)
////////////////////////
MC.world.afterEvents.entityHitBlock.subscribe(Event => {
    if (Event.damagingEntity.typeId === 'minecraft:player') {
        //
    };
});



// On Player Place Block // (after)
///////////////////////////
MC.world.afterEvents.playerPlaceBlock.subscribe(Event => {
    let placedPos = [Event.block.x, Event.block.y, Event.block.z];
    let foundGrid = false;
    for (let i = 0; i < 6; i++) {
        let placedOnPos = [Event.block.x, Event.block.y, Event.block.z];
        if (i === 0) { placedOnPos[1] += 1; }
        else if (i === 1) { placedOnPos[1] -= 1; }
        else if (i === 2) { placedOnPos[2] -= 1; }
        else if (i === 3) { placedOnPos[2] += 1; }
        else if (i === 4) { placedOnPos[0] -= 1; }
        else if (i === 5) { placedOnPos[0] += 1; };
        Object.keys(GRIDS.all).forEach(g => {
            let grid = GRIDS.all[g];
            let i_ = -1;
            GRIDS.all[g].areas.forEach(area => {
                i_ += 1;
                if (foundGrid === true) { return; };
                let result = PosInArea(area, placedPos);
                if (result === true) {
                    if (GridTerminalTypes.includes(Event.block.typeId) === true) {
                        grid.blocks.push({
                            'type': Event.block.typeId,
                            'position': placedPos,
                            'events': []
                        });
                    };
                    foundGrid = true;
                };
            });
        });
        if (foundGrid === true) { break; };
    };
    if (GridTerminalTypes.includes(Event.block.typeId) === true && foundGrid === false) {
        let gridId = NewGridId();
        let gridPos = [Event.block.x, Event.block.y, Event.block.z];
        GRIDS.all[gridId] = {
            'id': gridId,
            'position': gridPos,
            'dimension': Event.block.dimension.id,
            'velocity': [0, 0, 0],
            'blocks': [{
                'type': Event.block.typeId,
                'position': gridPos,
                'events': []
            }],
            'areas': [{
                'start': gridPos,
                'end': gridPos
            }],
            'config': { 'owner': null, 'gravity': false, 'drag': true, 'maxVelo': 1 }
        };
    };
});


MainTick();
//end